@@ -146,4 +146,4 @@ class VectorizerTester:
         return report
 
 
-#print(VectorizerTester('C:\workspace\pv_forecast\pvpower').report())
\ No newline at end of file
+print(VectorizerTester('C:\workspace\pv_forecast\pvpower').report())
\ No newline at end of file
